"""add-agent-version

Revision ID: 02950808ca3d
Revises: 4cc87e7fbfdf
Create Date: 2019-06-02 21:14:12.320029

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '02950808ca3d'
down_revision = '4cc87e7fbfdf'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'agents',
        # Set the defualt to "19.06.0" for now (since it is the first version to have this field)
        # and let the heartbeat handler route to update with the exact value.
        sa.Column('compute_plugins', postgresql.JSONB(astext_type=sa.Text()), nullable=False,
                  server_default=sa.text("'{}'::jsonb")))
    op.add_column(
        'agents',
        sa.Column('version', sa.String(length=64), nullable=False,
                  server_default=sa.literal('19.06.0')))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('agents', 'version')
    op.drop_column('agents', 'compute_plugins')
    # ### end Alembic commands ###



from __future__ import annotations

from typing import (
    Any,
    Optional,
    Sequence,
    Mapping,
)

import trafaret as t

from ai.backend.common.types import (
    AccessKey,
    AgentId,
    SessionId,
    ResourceSlot,
)

from .types import (
    AbstractScheduler,
    PendingSession,
    ExistingSession,
    AgentContext,
    KernelInfo,
)


class MOFScheduler(AbstractScheduler):
    """Minimum Occupied slot First Scheduler"""

    config_iv = t.Dict({}).allow_extra('*')

    def __init__(self, config: Mapping[str, Any]) -> None:
        super().__init__(config)

    def pick_session(
        self,
        total_capacity: ResourceSlot,
        pending_sessions: Sequence[PendingSession],
        existing_sessions: Sequence[ExistingSession],
    ) -> Optional[SessionId]:
        # Just pick the first pending session.
        return SessionId(pending_sessions[0].session_id)

    def _assign_agent(
        self,
        agents: Sequence[AgentContext],
        access_key: AccessKey,
        requested_slots: ResourceSlot,
    ) -> Optional[AgentId]:
        # return min occupied slot agent or None
        return next((one_agent.agent_id for one_agent in (sorted(
            (agent for agent in agents if (
                (agent.available_slots - agent.occupied_slots)
                >= requested_slots
            )),
            key=lambda a: a.occupied_slots)
        )), None)

    def assign_agent_for_session(
        self,
        agents: Sequence[AgentContext],
        pending_session: PendingSession,
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents, pending_session.access_key, pending_session.requested_slots,
        )

    def assign_agent_for_kernel(
        self,
        agents: Sequence[AgentContext],
        pending_kernel: KernelInfo,
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents, pending_kernel.access_key, pending_kernel.requested_slots,
        )



"""add-keypair-resource-policy

Revision ID: dbc1e053b880
Revises: 2b0931e4a059
Create Date: 2019-02-07 15:30:54.861821

"""
from alembic import op
import sqlalchemy as sa
from ai.backend.common.types import DefaultForUnspecified
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'dbc1e053b880'
down_revision = '2b0931e4a059'
branch_labels = None
depends_on = None


default_for_unspecified_choices = list(map(lambda v: v.name, DefaultForUnspecified))
default_for_unspecified = postgresql.ENUM(
    *default_for_unspecified_choices,
    name='default_for_unspecified',
)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'keypair_resource_policies',
        sa.Column('name', sa.String(length=256), nullable=False),
        sa.Column('created_at', sa.DateTime(timezone=True),
                  server_default=sa.text('now()'),
                  nullable=True),
        sa.Column('default_for_unspecified',
                  type_=sa.Enum(*default_for_unspecified_choices,
                                name='default_for_unspecified'),
                  nullable=False),
        sa.Column('total_resource_slots',
                  postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column('max_concurrent_sessions', sa.Integer(), nullable=False),
        sa.Column('max_containers_per_session', sa.Integer(), nullable=False),
        sa.Column('max_vfolder_count', sa.Integer(), nullable=False),
        sa.Column('max_vfolder_size', sa.BigInteger(), nullable=False),
        sa.Column('allowed_vfolder_hosts',
                  postgresql.ARRAY(sa.String()), nullable=False),
        sa.PrimaryKeyConstraint('name', name=op.f('pk_keypair_resource_policies'))
    )
    # Create a default resource policy
    query = '''
    INSERT INTO keypair_resource_policies
    VALUES (
        'default',
        now(),
        'UNLIMITED',
        '{}'::jsonb,
        30,
        1,
        10,
        0,
        array['local']
    );
    UPDATE keypairs
    SET resource_policy = 'default';
    '''
    connection = op.get_bind()
    connection.execute(query)
    print('\n!!! NOTICE !!!\n')
    print('Created a default resource policy and linked all keypairs to it.')
    print('Please inspect and adjust it!\n')
    op.alter_column('keypairs', 'resource_policy',
                    existing_type=sa.VARCHAR(),
                    type_=sa.String(length=256),
                    nullable=False)
    op.create_foreign_key(
        op.f('fk_keypairs_resource_policy_keypair_resource_policies'),
        'keypairs', 'keypair_resource_policies',
        ['resource_policy'], ['name'])
    op.drop_column('keypairs', 'concurrency_limit')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(op.f('fk_keypairs_resource_policy_keypair_resource_policies'),
                       'keypairs', type_='foreignkey')
    op.add_column(
        'keypairs',
        sa.Column('concurrency_limit', sa.INTEGER(),
                  autoincrement=False, nullable=True))
    op.alter_column('keypairs', 'resource_policy',
                    existing_type=sa.String(length=256),
                    type_=sa.VARCHAR(),
                    nullable=True)
    op.drop_table('keypair_resource_policies')
    default_for_unspecified.drop(op.get_bind())
    # ### end Alembic commands ###



"""add_user_group_ids_in_vfolder

Revision ID: 5d8e6043455e
Revises: 02950808ca3d
Create Date: 2019-06-06 15:02:58.804516

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql.expression import bindparam
from ai.backend.manager.models.base import convention, ForeignKeyIDColumn,  GUID, IDColumn


# revision identifiers, used by Alembic.
revision = '5d8e6043455e'
down_revision = '02950808ca3d'
branch_labels = None
depends_on = None


def upgrade():
    metadata = sa.MetaData(naming_convention=convention)
    # partial table to be preserved and referred
    keypairs = sa.Table(
        'keypairs', metadata,
        sa.Column('access_key', sa.String(length=20), primary_key=True),
        ForeignKeyIDColumn('user', 'users.uuid', nullable=False),
    )
    vfolders = sa.Table(
        'vfolders', metadata,
        IDColumn('id'),
        sa.Column('belongs_to', sa.String(length=20), sa.ForeignKey('keypairs.access_key')),
        sa.Column('user', GUID, sa.ForeignKey('users.uuid'), nullable=True),
        sa.Column('group', GUID, sa.ForeignKey('groups.id'), nullable=True),
    )
    vfolder_permissions = sa.Table(
        'vfolder_permissions', metadata,
        sa.Column('vfolder', GUID,
                  sa.ForeignKey('vfolders.id', onupdate='CASCADE', ondelete='CASCADE'),
                  nullable=False),
        sa.Column('access_key', sa.String(length=20), sa.ForeignKey('keypairs.access_key')),
        sa.Column('user', GUID, sa.ForeignKey('users.uuid'), nullable=True),
    )

    op.add_column('vfolders', sa.Column('user', GUID(), nullable=True))
    op.add_column('vfolders', sa.Column('group', GUID(), nullable=True))
    op.create_foreign_key(op.f('fk_vfolders_user_users'),
                          'vfolders', 'users', ['user'], ['uuid'])
    op.create_foreign_key(op.f('fk_vfolders_group_groups'),
                          'vfolders', 'groups', ['group'], ['id'])
    op.add_column('vfolder_permissions', sa.Column('user', GUID(), nullable=True))
    op.create_foreign_key(op.f('fk_vfolder_permissions_user_users'),
                          'vfolder_permissions', 'users', ['user'], ['uuid'])

    connection = op.get_bind()

    # Migrate vfolders' belongs_to keypair into user.
    j = vfolders.join(keypairs, vfolders.c.belongs_to == keypairs.c.access_key)
    query = sa.select([vfolders.c.id, keypairs.c.user]).select_from(j)
    results = connection.execute(query).fetchall()
    updates = [{'vid': row.id, 'user': row.user} for row in results]
    if updates:
        query = (sa.update(vfolders)
                   .values(user=bindparam('user'))
                   .where(vfolders.c.id == bindparam('vid')))
        connection.execute(query, updates)

    # Migrate vfolder_permissions' access_key into user.
    j = vfolder_permissions.join(keypairs,
                                 vfolder_permissions.c.access_key == keypairs.c.access_key)
    query = (sa.select([vfolder_permissions.c.vfolder, keypairs.c.access_key, keypairs.c.user])
               .select_from(j))
    results = connection.execute(query).fetchall()
    updates = [{'_vfolder': row.vfolder, '_access_key': row.access_key, '_user': row.user}
               for row in results]
    if updates:
        query = (sa.update(vfolder_permissions)
                   .values(user=bindparam('_user'))
                   .where(vfolder_permissions.c.vfolder == bindparam('_vfolder'))
                   .where(vfolder_permissions.c.access_key == bindparam('_access_key')))
        connection.execute(query, updates)

    op.drop_constraint('fk_vfolders_belongs_to_keypairs', 'vfolders', type_='foreignkey')
    op.drop_column('vfolders', 'belongs_to')
    op.alter_column('vfolder_permissions', 'user', nullable=False)
    op.drop_constraint('fk_vfolder_permissions_access_key_keypairs',
                       'vfolder_permissions', type_='foreignkey')
    op.drop_column('vfolder_permissions', 'access_key')


def downgrade():
    #######################################################################
    # CAUTION: group vfolders will be lost by downgrading this migration!
    #######################################################################

    metadata = sa.MetaData(naming_convention=convention)
    # partial table to be preserved and referred
    keypairs = sa.Table(
        'keypairs', metadata,
        sa.Column('access_key', sa.String(length=20), primary_key=True),
        ForeignKeyIDColumn('user', 'users.uuid', nullable=False),
    )
    vfolders = sa.Table(
        'vfolders', metadata,
        IDColumn('id'),
        sa.Column('belongs_to', sa.String(length=20), sa.ForeignKey('keypairs.access_key')),
        sa.Column('user', GUID, sa.ForeignKey('users.uuid'), nullable=True),
        # sa.Column('group', GUID, sa.ForeignKey('groups.id'), nullable=True),
    )
    vfolder_permissions = sa.Table(
        'vfolder_permissions', metadata,
        sa.Column('vfolder', GUID,
                  sa.ForeignKey('vfolders.id', onupdate='CASCADE', ondelete='CASCADE'),
                  nullable=False),
        sa.Column('access_key', sa.String(length=20), sa.ForeignKey('keypairs.access_key')),
        sa.Column('user', GUID, sa.ForeignKey('users.uuid'), nullable=True),
    )

    op.add_column('vfolders',
                  sa.Column('belongs_to', sa.String(length=20), autoincrement=False, nullable=True))
    op.create_foreign_key('fk_vfolders_belongs_to_keypairs',
                          'vfolders', 'keypairs', ['belongs_to'], ['access_key'])
    op.add_column('vfolder_permissions',
                  sa.Column('access_key', sa.String(length=20), autoincrement=False, nullable=True))
    op.create_foreign_key('fk_vfolder_permissions_access_key_keypairs',
                          'vfolder_permissions', 'keypairs', ['access_key'], ['access_key'])

    connection = op.get_bind()

    # Migrate vfolders' user_id into belongs_to.
    j = vfolders.join(keypairs, vfolders.c.user == keypairs.c.user)
    query = sa.select([vfolders.c.id, keypairs.c.access_key]).select_from(j)
    results = connection.execute(query).fetchall()
    updates = [{'vid': row.id, 'belongs_to': row.access_key} for row in results]
    if updates:
        query = (sa.update(vfolders)
                   .values(belongs_to=bindparam('belongs_to'))
                   .where(vfolders.c.id == bindparam('vid')))
        connection.execute(query, updates)

    # Migrate vfolder_permissions' used into access_key.
    j = (vfolder_permissions.join(keypairs, vfolder_permissions.c.user == keypairs.c.user))
    query = (sa.select([vfolder_permissions.c.vfolder, keypairs.c.user, keypairs.c.access_key])
               .select_from(j))
    results = connection.execute(query).fetchall()
    updates = [{'_vfolder': row.vfolder, '_access_key': row.access_key, '_user': row.user} \
               for row in results]
    if updates:
        query = (sa.update(vfolder_permissions)
                   .values(access_key=bindparam('_access_key'))
                   .where(vfolder_permissions.c.vfolder == bindparam('_vfolder'))
                   .where(vfolder_permissions.c.user == bindparam('_user')))
        connection.execute(query, updates)

    op.alter_column('vfolders', 'belongs_to', nullable=False)
    op.alter_column('vfolder_permissions', 'access_key', nullable=False)
    op.drop_constraint(op.f('fk_vfolders_user_users'), 'vfolders', type_='foreignkey')
    op.drop_constraint(op.f('fk_vfolders_group_groups'), 'vfolders', type_='foreignkey')
    op.drop_column('vfolders', 'user')
    op.drop_column('vfolders', 'group')
    op.drop_constraint(op.f('fk_vfolder_permissions_user_users'), 'vfolder_permissions', type_='foreignkey')
    op.drop_column('vfolder_permissions', 'user')



"""change-kernel-identification

Revision ID: 854bd902b1bc
Revises: 0f3bc98edaa0
Create Date: 2017-08-21 17:08:20.581565

"""
from alembic import op
import sqlalchemy as sa
from ai.backend.manager.models.base import GUID
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '854bd902b1bc'
down_revision = '0f3bc98edaa0'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint('fk_vfolder_attachment_vfolder_vfolders', 'vfolder_attachment', type_='foreignkey')
    op.drop_constraint('fk_vfolder_attachment_kernel_kernels', 'vfolder_attachment', type_='foreignkey')
    op.drop_constraint('pk_kernels', 'kernels', type_='primary')
    op.add_column('kernels',
                  sa.Column('id', GUID(),
                            server_default=sa.text('uuid_generate_v4()'),
                            nullable=False))
    op.add_column('kernels', sa.Column('role', sa.String(length=16), nullable=False, default='master'))
    op.create_primary_key('pk_kernels', 'kernels', ['id'])
    op.alter_column(
        'kernels', 'sess_id',
        existing_type=postgresql.UUID(),
        type_=sa.String(length=64),
        nullable=True,
        existing_server_default=sa.text('uuid_generate_v4()'))
    op.create_index(op.f('ix_kernels_sess_id'), 'kernels', ['sess_id'], unique=False)
    op.create_index(op.f('ix_kernels_sess_id_role'), 'kernels', ['sess_id', 'role'], unique=False)
    op.create_foreign_key('fk_vfolder_attachment_vfolder_vfolders',
                          'vfolder_attachment', 'vfolders',
                          ['vfolder'], ['id'], onupdate='CASCADE', ondelete='CASCADE')
    op.create_foreign_key('fk_vfolder_attachment_kernel_kernels',
                          'vfolder_attachment', 'kernels',
                          ['kernel'], ['id'], onupdate='CASCADE', ondelete='CASCADE')


def downgrade():
    op.drop_constraint('fk_vfolder_attachment_vfolder_vfolders', 'vfolder_attachment', type_='foreignkey')
    op.drop_constraint('fk_vfolder_attachment_kernel_kernels', 'vfolder_attachment', type_='foreignkey')
    op.drop_constraint('pk_kernels', 'kernels', type_='primary')
    op.drop_index(op.f('ix_kernels_sess_id'), table_name='kernels')
    op.drop_index(op.f('ix_kernels_sess_id_role'), table_name='kernels')
    op.alter_column(
        'kernels', 'sess_id',
        existing_type=sa.String(length=64),
        type_=postgresql.UUID(),
        nullable=False,
        existing_server_default=sa.text('uuid_generate_v4()'),
        postgresql_using='sess_id::uuid')
    op.create_primary_key('pk_kernels', 'kernels', ['sess_id'])
    op.drop_column('kernels', 'id')
    op.drop_column('kernels', 'role')
    op.create_foreign_key('fk_vfolder_attachment_vfolder_vfolders',
                          'vfolder_attachment', 'vfolders',
                          ['vfolder'], ['id'])
    op.create_foreign_key('fk_vfolder_attachment_kernel_kernels',
                          'vfolder_attachment', 'kernels',
                          ['kernel'], ['sess_id'])
